from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Generator
from typing import List

import numpy as np
import pandas as pd
import yaml

from pynear.benchmark.dataset import generate_gaussian_dataset
from pynear.benchmark.index_adapters import create_index_adapter
from pynear.logging import create_and_configure_log

logger = create_and_configure_log(__name__)

# number of searchs to perform for averaging the result
# in order to reduce effect of high outliers
NUM_AVG_SEARCHS = 8


@dataclass
class BenchmarkCase:
    """
    A benchmark case to run.
    """

    name: str

    # list of test case K values (how many neighbors to search)
    # each case will result in multiple tests, one for each k value
    k: List[int]
    num_queries: List[int]
    index_types: List[str]
    dimensions: List[int]
    dataset_total_size: int
    dataset_num_clusters: int
    dataset_type: str = "float32"

    def run(self):
        results = []
        for dimension in self.dimensions:
            logger.info("generating dataset ...")
            data = generate_gaussian_dataset(
                self.dataset_total_size,
                self.dataset_num_clusters,
                dimension,
                data_type=np.dtype(self.dataset_type),
            )
            logger.info("generating dataset done")
            for k_value in self.k:
                for index_type in self.index_types:
                    logger.info(f"processing index_type: {index_type} for k = {k_value} and dimension = {dimension}")
                    index = create_index_adapter(index_type)
                    logger.info("buiding index ...")
                    index.build_index(data)
                    logger.info("buiding index done")
                    logger.info("start performing queries")
                    for num_queries in self.num_queries:
                        query = generate_gaussian_dataset(
                            num_queries,
                            1,
                            dimension,
                            data_type=np.dtype(self.dataset_type),
                        )
                        logger.info(f"start performing queries (num_queries = {num_queries})")
                        runs = np.array([index.clock_search(query, k_value) for _ in range(NUM_AVG_SEARCHS)])
                        logger.info(f"5 runs set: {runs}")
                        runs = BenchmarkCase.reject_outliers(runs)
                        logger.info(f"rejected {NUM_AVG_SEARCHS-len(runs)} outliers.")

                        # reject outliers
                        avg = sum(runs) / len(runs)
                        logger.info(f"avg of {NUM_AVG_SEARCHS} searches runtime is {avg:0.4f}")

                        results.append(
                            {
                                "time": avg,
                                "k": k_value,
                                "num_seraches_avg": NUM_AVG_SEARCHS,
                                "num_queries": num_queries,
                                "index_type": index_type,
                                "dimension": dimension,
                                "dataset_total_size": self.dataset_total_size,
                                "dataset_num_clusters": self.dataset_num_clusters,
                            }
                        )
                        logger.info(f"done performing queries for (num_queries = {num_queries})")
                    logger.info("queries done")
        return {"benchmark_case_id": self.id(), "benchmark_case_name": self.name, "results": results}

    # the random seed for reproducibility
    seed: int = 12246

    def id(self):
        # return name in a dns compatible format
        return self.name.replace(" ", "-").lower()

    def __str__(self):
        return f"{self.id()}: ks={self.k}, num_queries={self.num_queries}, index_types={self.index_types}, dataset_total_size={self.dataset_total_size}, num_clusters={self.dataset_num_clusters}, seed={self.seed}"

    @staticmethod
    def reject_outliers(data, m=1.2):
        return data[abs(data - np.mean(data)) < m * np.std(data)]


class BenchmarkRunner:
    """
    A generic benchmark build helper class to automate benchmark generation.
    """

    def __init__(self, benchmark_yaml_file: str):
        """
        Build a benchmark comparator between pynear and faiss.

        Args:
            benchmark_cases (List[BenchmarkCase]): the list of benchmark cases to perform.
        """
        self._benchmark_cases = BenchmarkRunner.read_cases_from_yaml(benchmark_yaml_file)
        self._results = pd.DataFrame()

    @staticmethod
    def read_cases_from_yaml(yaml_file: str) -> List[BenchmarkCase]:
        """
        Read benchmark cases from a yaml file.

        Args:
            yaml_file (str): the yaml file to read from.
        """
        with open(yaml_file, "r") as stream:
            benchmark_configs = yaml.safe_load(stream)

        cases = benchmark_configs.get("benchmark", {}).get("cases", [])

        results = []
        for case in cases:
            results.append(BenchmarkCase(**case))

        return results

    def run(self) -> Generator[Dict[str, Any], None, None]:
        num_cases = len(self._benchmark_cases)
        logger.info(f"start running benchmark for {num_cases} cases ...")

        for case in self._benchmark_cases:
            logger.info(f"***** begin case {case.name} *****\n")
            case_result = case.run()
            logger.info(f"******* end case {case.name} *****\n")
            yield case_result



import numpy as np
import pytest

import pynear

CLASSES = [
    (pynear.BKTreeBinaryIndex64, 8),
    (pynear.BKTreeBinaryIndex128, 16),
    (pynear.BKTreeBinaryIndex256, 32),
    (pynear.BKTreeBinaryIndex512, 64),
    (pynear.BKTreeBinaryIndexN, 1),
]


def hamming_distance_pairwise(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    r = (1 << np.arange(8))[:, None, None, None]
    return np.count_nonzero((np.bitwise_xor(a[:, None, :], b[None, :, :]) & r) != 0, axis=(0, -1))


@pytest.mark.parametrize("bktree_cls, dimensions", CLASSES)
def test_bktree_empty_index(bktree_cls, dimensions):
    num_points = 2
    data = np.random.randint(0, 255, size=(num_points, dimensions), dtype=np.uint8)
    empty = np.array([], dtype=np.uint8)

    tree = bktree_cls()
    indices, distances, keys = tree.find_threshold(data, 1)
    truth = [[]] * num_points
    assert truth == indices
    assert truth == distances
    assert tree.empty()
    assert tree.values() == []

    tree.set(empty)
    indices, distances, keys = tree.find_threshold(data, 1)
    truth = [[]] * num_points
    assert truth == indices
    assert truth == distances
    assert tree.empty()
    assert tree.values() == []


@pytest.mark.parametrize("bktree_cls, dimensions", CLASSES)
def test_bktree_find_self(bktree_cls, dimensions):
    num_points = 2
    data = np.random.randint(0, 255, size=(num_points, dimensions), dtype=np.uint8)

    tree = bktree_cls()
    tree.set(data)
    indices, distances, keys = tree.find_threshold(data, 0)
    assert indices == [[i] for i in range(num_points)]
    assert distances == [[0]] * num_points
    assert keys == data[:, None, :].tolist()
    assert tree.size() == num_points
    assert sorted(tree.values()) == sorted(data.tolist())


@pytest.mark.parametrize("bktree_cls, dimensions", CLASSES)
def test_bktree_find_all(bktree_cls, dimensions):
    num_points = 2
    data = np.random.randint(0, 255, size=(num_points, dimensions), dtype=np.uint8)

    tree = bktree_cls()
    tree.set(data)
    indices, distances, keys = tree.find_threshold(data, dimensions * 8)

    assert indices == [list(range(num_points))] * num_points
    assert distances == hamming_distance_pairwise(data, data).tolist()
    assert keys == np.broadcast_to(data, (num_points, num_points, dimensions)).tolist()
    assert tree.size() == num_points
    assert sorted(tree.values()) == sorted(data.tolist())


@pytest.mark.parametrize("bktree_cls, dimensions", CLASSES)
def test_bktree_find_duplicates(bktree_cls, dimensions):
    num_points = 2
    data = np.zeros((num_points, dimensions), dtype=np.uint8)

    tree = bktree_cls()
    tree.set(data)
    indices, distances, keys = tree.find_threshold(data, 255)

    assert indices == [list(range(num_points))] * num_points
    assert distances == [[0] * num_points] * num_points
    assert keys == np.broadcast_to(data, (num_points, num_points, dimensions)).tolist()
    assert tree.size() == num_points
    assert sorted(tree.values()) == sorted(data.tolist())



from abc import ABC
from abc import abstractmethod
import time

import annoy
import faiss
import numpy as np
from sklearn.neighbors import NearestNeighbors

import pynear


def create_index_adapter(index_name: str):
    # Supported 3rd party indices are: FaissIndexFlatL2, FaissIndexBinaryFlat, AnnoyL2, AnnoyManhattan, AnnoyHamming, SKLearnL2
    mapper = {
        "FaissIndexFlatL2": FaissIndexFlatL2Adapter,
        "FaissIndexBinaryFlat": FaissIndexBinaryFlatAdapter,
        "AnnoyL2": AnnoyL2Adapter,
        "AnnoyManhattan": AnnoyManhattanAdapter,
        "AnnoyHamming": AnnoyHammingAdapter,
        "SKLearnL2": SKLearnL2Adapter,
        "BKTreeBinaryIndex": PyNearBKTreeAdapter,
        "VPTreeL2Index": pynear.VPTreeL2Index,
        "VPTreeL1Index": pynear.VPTreeL1Index,
        "VPTreeBinaryIndex": pynear.VPTreeBinaryIndex,
        "VPTreeChebyshevIndex": pynear.VPTreeChebyshevIndex,
    }
    if index_name not in mapper:
        raise ValueError(f"Index name {index_name} not supported")

    if index_name.startswith("VPTree"):
        return PyNearVPAdapter(index_name)

    return mapper[index_name]()


# Supported 3rd party indices are: FaissIndexFlatL2, FaissIndexBinaryFlat, AnnoyL2, AnnoyManhattan, AnnoyHamming, SKLearnL2
class IndexAdapter(ABC):
    @abstractmethod
    def build_index(self, data: np.ndarray):
        pass

    def clock_search(self, query: np.ndarray, k: int) -> float:
        """
        Searchs in index and retrieves the time it took to search
        """
        s = time.time()
        self._search_implementation(query, k)
        return time.time() - s

    @abstractmethod
    def _search_implementation(self, query, k: int):
        pass


class PyNearVPAdapter(IndexAdapter):
    def __init__(self, pyvp_index_name: str):
        self._index = None
        self._pyvp_index_name = pyvp_index_name
        self._pyvp_index_map = {
            "VPTreeL2Index": pynear.VPTreeL2Index,
            "VPTreeBinaryIndex": pynear.VPTreeBinaryIndex,
            "VPTreeChebyshevIndex": pynear.VPTreeChebyshevIndex,
            "VPTreeL1Index": pynear.VPTreeL1Index,
        }

    def build_index(self, data: np.ndarray):
        self._index = self._pyvp_index_map[self._pyvp_index_name]()
        self._index.set(data)

    def _search_implementation(self, query, k: int):
        self._index.searchKNN(query, k)


class PyNearBKTreeAdapter(IndexAdapter):
    def __init__(self):
        self._index = pynear.BKTreeBinaryIndex()
        self._dimensions = 0

    def build_index(self, data: np.ndarray):
        self._index.set(data)
        self._dimensions = data.shape[1]

    def _search_implementation(self, query, k: int):
        self._index.find_threshold(query, self._dimensions)


class FaissIndexFlatL2Adapter(IndexAdapter):
    def __init__(self):
        self._index = None

    def build_index(self, data: np.ndarray):
        d = data.shape[1]
        self._index = faiss.IndexFlatL2(d)
        self._index.add(data)

    def _search_implementation(self, query, k: int):
        self._index.search(query, k=k)


class FaissIndexBinaryFlatAdapter(IndexAdapter):
    def __init__(self):
        self._index = None

    def build_index(self, data: np.ndarray):
        d = data.shape[1] * 8
        self._index = faiss.IndexBinaryFlat(d)
        self._index.add(data)

    def _search_implementation(self, query, k: int):
        self._index.search(query, k=k)


class AnnoyL2Adapter(IndexAdapter):
    def __init__(self):
        self._index = None
        pass

    def build_index(self, data: np.ndarray):
        self._index = annoy.AnnoyIndex(data.shape[1], "euclidean")
        for i, v in enumerate(data):
            self._index.add_item(i, v)

        self._index.build(10)

    def _search_implementation(self, query, k: int):
        for v in query:
            self._index.get_nns_by_vector(v, k)


class AnnoyManhattanAdapter(IndexAdapter):
    def __init__(self):
        self._index = None

    def build_index(self, data: np.ndarray):
        self._index = annoy.AnnoyIndex(data.shape[1], "manhattan")
        for i, v in enumerate(data):
            self._index.add_item(i, v)

    def _search_implementation(self, query, k: int):
        for v in query:
            self._index.get_nns_by_vector(v, k)


class AnnoyHammingAdapter(IndexAdapter):
    def __init__(self):
        self._index = None

    def build_index(self, data: np.ndarray):
        self._index = annoy.AnnoyIndex(data.shape[1], "hamming")
        for i, v in enumerate(data):
            self._index.add_item(i, v)

        self._index.build(10)

    def _search_implementation(self, query, k: int):
        for v in query:
            self._index.get_nns_by_vector(v, k)


class SKLearnL2Adapter(IndexAdapter):
    def __init__(self):
        self._index = None

    def build_index(self, data: np.ndarray):
        self._data = data
        pass

    def _search_implementation(self, query, k: int):
        # sklearn uses index based on k, so need to build on the fly
        pass

    def clock_search(self, query: np.ndarray, k: int) -> float:
        """
        Searchs in index and retrieves the time it took to search
        """
        # sklearn uses index based on k, so need to build on the fly
        # we will not clock index training stage
        self._index = NearestNeighbors(n_neighbors=k, algorithm="kd_tree", metric="euclidean")
        self._index.fit(self._data)

        s = time.time()
        self._index.kneighbors(query)
        return time.time() - s



import argparse
import os

import matplotlib.pyplot as plt
import pandas as pd

from pynear.benchmark import BenchmarkRunner
from pynear.logging import create_and_configure_log

logger = create_and_configure_log(__name__)


def create_performance_plot(result: pd.DataFrame, case_name: str, output_folder: str):
    groups = result.groupby("k")

    for k, group in groups:
        index_types = group.groupby("index_type")
        for index_type, df in index_types:
            # create one plot per k
            # resetting index before melting to save the current index in 'index' column...
            plt.plot([str(v) for v in df["dimension"]], df["time"], label=index_type, marker="o")
            data_size = df.iloc[0]["dataset_total_size"]
            query_size = df.iloc[0]["num_queries"]
            num_avg_searchs = df.iloc[0]["num_seraches_avg"]
            plt.title(
                f"{case_name}\nData Dimensions x Query Time (avg of {num_avg_searchs})\ndata_size={data_size}), k={k}\nquery_size={query_size})"
            )
            plt.legend(loc="upper center")

        plt.tight_layout()
        plt.savefig(os.path.join(output_folder, f"result_k={k}.png"))
        plt.clf()


def main():
    parser = argparse.ArgumentParser(description="Creates a stress test report for a segemaker endpoint")
    parser.add_argument(
        "--config-file",
        default="benchmark_config.yaml",
        type=str,
        help="The config file describing the benchmark cases to run",
        required=False,
    )

    args = parser.parse_args()

    runner = BenchmarkRunner(args.config_file)

    print("start running benchmarks... ")

    for result in runner.run():
        case_id = result["benchmark_case_id"]
        case_name = result["benchmark_case_name"]
        case_results = result["results"]
        dir_name = f"./results/{case_id}"
        os.makedirs(dir_name, exist_ok=True)
        create_performance_plot(pd.DataFrame(case_results), case_name, dir_name)

    print("benchmarks done.")
    print("result images written to ./results folder")


if __name__ == "__main__":
    main()



from typing import Any

import numpy as np


def generate_gaussian_dataset(total_size: int, num_clusters: int, dim: int, data_type: Any = np.float64) -> np.ndarray:
    """
    Generate a set of gaussian clusters of specific size and dimention
    """
    # sample gaussan center points
    cluster_size = total_size // num_clusters

    centers = None
    scalers = None
    if data_type == np.uint8:
        centers = np.random.uniform(0, 255, size=(cluster_size, dim))
        scalers = np.random.uniform(1, 1, size=(cluster_size, dim))
    else:
        centers = np.random.uniform(-10000, 10000, size=(cluster_size, dim))
        scalers = np.random.uniform(1, 10, size=(cluster_size, dim))
    datas: list = []
    for cluster in range(num_clusters):
        center = centers[cluster, :]
        scale = scalers[cluster, :]
        data = np.random.normal(loc=center, scale=scale, size=(cluster_size, dim))
        datas.append(data)

    return np.concatenate(datas, axis=0).astype(data_type)



from typing import List
from typing import Tuple

import numpy as np

from _pynear import BKTreeBinaryIndex64
from _pynear import BKTreeBinaryIndex128
from _pynear import BKTreeBinaryIndex256
from _pynear import BKTreeBinaryIndex512
from _pynear import BKTreeBinaryIndex as BKTreeBinaryIndexN
from _pynear import VPTreeBinaryIndex64
from _pynear import VPTreeBinaryIndex128
from _pynear import VPTreeBinaryIndex256
from _pynear import VPTreeBinaryIndex512
from _pynear import VPTreeBinaryIndex as VPTreeBinaryIndexN
from _pynear import VPTreeChebyshevIndex
from _pynear import VPTreeL1Index
from _pynear import VPTreeL2Index
from _pynear import dist_chebyshev
from _pynear import dist_hamming_64
from _pynear import dist_hamming_128
from _pynear import dist_hamming_256
from _pynear import dist_hamming_512
from _pynear import dist_l1
from _pynear import dist_l2

from ._version import __version__


def dist_hamming(a: List, b: List):
    if len(a) != len(b):
        raise ValueError("invalid data dimension: a and b dimensions must agree.")
    dim = len(a)
    if dim == 64:
        return dist_hamming_512(a, b)
    elif dim == 32:
        return dist_hamming_256(a, b)
    elif dim == 16:
        return dist_hamming_128(a, b)
    elif dim == 8:
        return dist_hamming_64(a, b)
    else:
        raise ValueError("invalid data dimension: hamming distance only supports 64, 32, 16 or 8 bytes of data")


class VPTreeBinaryIndex:
    def __init__(self) -> None:
        self._index = None
        self._dimension = None

    def set(self, data: np.ndarray) -> None:
        self._validate(data)

        dim = data.shape[1]
        if dim == 64:
            self._index = VPTreeBinaryIndex512()
        elif dim == 32:
            self._index = VPTreeBinaryIndex256()
        elif dim == 16:
            self._index = VPTreeBinaryIndex128()
        elif dim == 8:
            self._index = VPTreeBinaryIndex64()
        else:
            self._index = VPTreeBinaryIndexN()

        self._dimension = dim
        self._index.set(data)

    def searchKNN(self, queries: np.ndarray, k: int) -> Tuple[list, list]:
        dim = queries.shape[1]
        if dim != self._dimension:
            raise ValueError(
                f"invalid data dimension: index built data and query data dimensions must agree, index built data dimension is {dim}"
            )

        if self._index is None:
            return [], []

        self._validate(queries)
        return self._index.searchKNN(queries, k)

    def search1NN(self, queries: np.ndarray) -> Tuple[list, list]:
        if self._index is None:
            return [], []

        self._validate(queries)
        return self._index.search1NN(queries)

    def _validate(self, data: np.ndarray) -> None:
        if len(data.shape) != 2:
            raise ValueError("invalid data shape: binary indexes must be 2D")

        if data.dtype != "uint8":
            raise TypeError("invalid data type: binary indexes must be uint8")


class BKTreeBinaryIndex:
    def __init__(self) -> None:
        self._index = None
        self._dimension = None

    def set(self, data: np.ndarray) -> None:
        self._validate(data)

        dim = data.shape[1]
        if dim == 64:
            self._index = BKTreeBinaryIndex512()
        elif dim == 32:
            self._index = BKTreeBinaryIndex256()
        elif dim == 16:
            self._index = BKTreeBinaryIndex128()
        elif dim == 8:
            self._index = BKTreeBinaryIndex64()
        else:
            self._index = BKTreeBinaryIndexN()

        self._dimension = dim
        self._index.set(data)

    def find_threshold(self, queries: np.ndarray, threshold: int) -> Tuple[list, list]:
        dim = queries.shape[1]
        if dim != self._dimension:
            raise ValueError(
                f"invalid data dimension: index built data and query data dimensions must agree, index built data dimension is {dim}"
            )

        if self._index is None:
            return [], []

        self._validate(queries)
        return self._index.find_threshold(queries, threshold)

    def empty(self) -> bool:
        if self._index is None:
            return True

        return self._index.empty()

    def values(self) -> list:
        if self._index is None:
            return []

        return self._index.values()

    def _validate(self, data: np.ndarray) -> None:
        if len(data.shape) != 2:
            raise ValueError("invalid data shape: binary indexes must be 2D")

        if data.dtype != "uint8":
            raise TypeError("invalid data type: binary indexes must be uint8")



import logging
import os


def create_and_configure_log(name):
    logger = logging.getLogger(name)
    logger.setLevel("INFO")
    level_envar = "PYNEAR_LOG_LEVEL"
    if level_envar in os.environ:
        logger.setLevel(os.environ[level_envar])

    # TODO: proper logging configuration
    formatter = logging.Formatter("%(asctime)s %(processName)-10s %(name)s %(levelname)-8s %(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    return logger



import pickle

import numpy as np

import pynear


def test_empty_index_serialization():
    vptree = pynear.VPTreeL2Index()
    data = pickle.dumps(vptree)
    recovered = pickle.loads(data)
    data_rec = pickle.dumps(recovered)
    assert data_rec == data

    # not initializing data with .set() should
    # be equivalent of initializing with empty array
    vptree = pynear.VPTreeL2Index()
    empty = np.array([]).reshape(-1, 8)
    vptree.set(empty)
    data = pickle.dumps(vptree)
    recovered = pickle.loads(data)
    data_rec = pickle.dumps(recovered)
    assert data_rec == data

    vptree = pynear.VPTreeBinaryIndex()
    data = pickle.dumps(vptree)
    recovered = pickle.loads(data)
    data_rec = pickle.dumps(recovered)
    assert data_rec == data


def test_basic_serialization():
    num_points = 10
    dimension = 8
    num_queries = 2
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    vptree = pynear.VPTreeL2Index()
    vptree.set(data)

    vptree_indices, vptree_distances = vptree.search1NN(queries)

    data = pickle.dumps(vptree)
    recovered = pickle.loads(data)
    recovered_data = pickle.dumps(recovered)

    assert data == recovered_data

    vptree_indices_rec, vptree_distances_rec = recovered.search1NN(queries)
    assert vptree_indices_rec == vptree_indices and vptree_distances_rec == vptree_distances


def test_string_serialization():
    num_points = 27
    dimension = 8
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    vptree = pynear.VPTreeL2Index()
    vptree.set(data)

    string_state = vptree.to_string()

    binary_state = pickle.dumps(vptree)
    restored = pickle.loads(binary_state)

    restored_string_state = restored.to_string()

    assert restored_string_state == string_state


def test_binary_serialization():
    num_points = 20000
    dimension = 32
    num_queries = 2
    data = np.random.rand(num_points, dimension).astype(dtype=np.uint8)

    queries = np.random.rand(num_queries, dimension).astype(dtype=np.uint8)

    vptree = pynear.VPTreeBinaryIndex()
    vptree.set(data)

    vptree_indices, vptree_distances = vptree.search1NN(queries)

    data = pickle.dumps(vptree)
    recovered = pickle.loads(data)
    recovered_data = pickle.dumps(recovered)

    assert data == recovered_data

    vptree_indices_rec, vptree_distances_rec = recovered.search1NN(queries)
    assert vptree_distances_rec == vptree_distances



import sys

from pybind11.setup_helpers import Pybind11Extension
from setuptools import find_packages
from setuptools import setup

if sys.platform == "win32":
    extra_compile_args = ["/Wall", "/arch:AVX", "/openmp"]  # /LTCG unrecognized here
    extra_link_args = ["/LTCG"]  # /openmp unrecognized here
elif sys.platform == "darwin":
    extra_compile_args = ["-flto", "-Wall", "-march=native", "-mavx", "-fopenmp"]
    extra_link_args = ["-fopenmp", "-lomp"]
else:
    extra_compile_args = ["-flto", "-Wall", "-march=native", "-mavx", "-fopenmp"]
    extra_link_args = ["-fopenmp", "-lgomp"]

ext_modules = [
    Pybind11Extension(
        "_pynear",
        ["pynear/src/PythonBindings.cpp"],
        include_dirs=["pynear/include"],
        cxx_std=17,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        define_macros=[("ENABLE_OMP_PARALLEL", "1")],
    ),
]

with open("README.md", "rt", encoding="utf-8") as fr:
    long_description = fr.read()

exec(open("pynear/_version.py").read())
setup(
    name="pynear",
    version=__version__,  # noqa: F821
    packages=find_packages(),
    author="Pablo Carneiro Elias",
    author_email="pablo.cael@gmail.com",
    url="https://github.com/pablocael/pynear",
    description="An efficient implementation of Vantage Point Tree for Python 3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    zip_safe=False,
    install_requires=["numpy>=1.21.2"],
    package_dir={"pynear": "pynear"},
    extras_require={"test": ["pytest>=6.0"]},
    python_requires=">=3.7",
    license_files=("LICENSE",),
)



#
# MIT Licence
# Copyright 2021 Pablo Carneiro Elias
#

from collections import Counter
from functools import partial
import os
import sys
from typing import Callable
from typing import Tuple

import numpy as np
import pytest

import pynear
from pynear import dist_chebyshev
from pynear import dist_hamming
from pynear import dist_l1
from pynear import dist_l2

seed = os.environ.get("PYNEAR_TEST_SEED", None)
if seed is not None:
    np.random.seed(int(seed))
    print(f"using test seed: {seed}")

np.set_printoptions(threshold=sys.maxsize)


def hamming_distance_pairwise(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    result = []
    for ai in a:
        dists = []
        for bi in b:
            dists.append(dist_hamming(ai, bi))

        result.append(dists)

    return np.array(result).astype(np.uint8)


def euclidean_distance_pairwise(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    result = []
    for ai in a:
        dists = []
        for bi in b:
            dists.append(dist_l2(ai, bi))

        result.append(dists)

    return np.array(result)


def manhattan_distance_pairwise(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    result = []
    for ai in a:
        dists = []
        for bi in b:
            dists.append(dist_l1(ai, bi))

        result.append(dists)

    return np.array(result)


def chebyshev_distance_pairwise(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    result = []
    for ai in a:
        dists = []
        for bi in b:
            dists.append(dist_chebyshev(ai, bi))

        result.append(dists)

    return np.array(result)


def test_empty_index():
    try:
        vptree = pynear.VPTreeBinaryIndex()
        i, d = vptree.search1NN(np.random.rand(1, 8).astype(dtype=np.uint8))
        # expect exception and should now reach this line
        # since index is empty, we should not be able to search
        assert False
    except Exception:
        pass

    try:
        vptree = pynear.VPTreeL2Index()
        i, d = vptree.search1NN(np.random.rand(1, 8).astype(dtype=np.uint8))
        # expect exception and should now reach this line
        # since index is empty, we should not be able to search
        assert False
    except Exception:
        pass

    try:
        vptree = pynear.VPTreeL2Index()
        empty = np.array([]).reshape(-1, 2)
        vptree.set(empty)
        i, d = vptree.search1NN(np.random.rand(1, 8).astype(dtype=np.uint8))
        assert False

        # expect no exception and should now reach this line
        # since index is empty, we should not be able to search
    except Exception:
        pass


def test_hamming():
    def hamming_distance(a, b) -> np.ndarray:
        r = (1 << np.arange(8))[:, None]
        return np.count_nonzero((np.bitwise_xor(a, b) & r) != 0)

    arr1 = np.random.randint(0, 10, (5, 8), dtype=np.uint8)
    arr2 = np.random.randint(0, 10, (3, 8), dtype=np.uint8)

    truth = np.empty((arr1.shape[0], arr2.shape[0]), dtype=np.uint64)
    for i in range(arr1.shape[0]):
        for j in range(arr2.shape[0]):
            truth[i, j] = hamming_distance(arr1[i], arr2[j])

    result = hamming_distance_pairwise(arr1, arr2)

    assert np.array_equal(truth, result)


def exhaustive_search(
    metric_func: Callable[[np.ndarray, np.ndarray], np.ndarray],
    data: np.ndarray,
    queries: np.ndarray,
    k: int,
) -> Tuple[np.ndarray, np.ndarray]:
    distances = metric_func(queries, data)
    indices = np.argpartition(distances, range(k), axis=-1)[:, :k]
    distances = np.take_along_axis(distances, indices, axis=-1)

    return indices, distances


exhaustive_search_hamming = partial(exhaustive_search, hamming_distance_pairwise)
exhaustive_search_euclidean = partial(exhaustive_search, euclidean_distance_pairwise)
exhaustive_search_manhattan = partial(exhaustive_search, manhattan_distance_pairwise)
exhaustive_search_chebyshev = partial(exhaustive_search, chebyshev_distance_pairwise)


def _num_dups(distances):
    dups = 0
    for i in range(len(distances)):
        c = Counter(distances[i].tolist())
        dups += sum(1 for k, c in c.most_common() if c > 1)
    return dups


CLASSES = [
    (pynear.VPTreeL2Index, exhaustive_search_euclidean),
    (pynear.VPTreeL1Index, exhaustive_search_manhattan),
    (pynear.VPTreeChebyshevIndex, exhaustive_search_chebyshev),
]


@pytest.mark.parametrize("num_points, k", [(2021, 2), (40021, 3)])
def test_binary(num_points, k):
    dimension = 32
    data = np.random.normal(scale=255, loc=0, size=(num_points, dimension)).astype(dtype=np.uint8)

    num_queries = 8
    queries = np.random.normal(scale=255, loc=0, size=(num_queries, dimension)).astype(dtype=np.uint8)

    exaustive_indices, exaustive_distances = exhaustive_search_hamming(data, queries, k)

    vptree = pynear.VPTreeBinaryIndex()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.int64)[:, ::-1]

    assert np.array_equal(exaustive_distances, vptree_distances)
    # assert np.array_equal(exaustive_indices, vptree_indices)  # indices order can vary for same distances


def test_binary_duplicates():
    dimension = 32
    num_points = 2
    data = np.zeros((num_points, dimension), dtype=np.uint8)

    k = 2

    vptree = pynear.VPTreeBinaryIndex()
    vptree.set(data)
    indices, distances = vptree.searchKNN(data, k)

    assert [sorted(i) for i in indices] == [list(range(num_points))] * num_points
    assert distances == [[0] * k] * num_points


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_k_equals_dataset(vptree_cls, exaustive_metric):
    dimension = 8
    num_points = 2021
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    num_queries = 8
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    k = num_points

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float64)[:, ::-1]
    dist_diff = vptree_distances - exaustive_distances
    ind_diff = vptree_indices - exaustive_indices
    print(">>>>>>>>>>>>", dist_diff[dist_diff > 1e-7])
    print(">>>>>>>>>>>>", np.argwhere(ind_diff != 0))

    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)
    if _num_dups(exaustive_distances) == 0:
        assert np.array_equal(exaustive_indices, vptree_indices)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_large_dataset(vptree_cls, exaustive_metric):
    dimension = 8
    num_points = 401001
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    num_queries = 8
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    k = 3

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float32)[:, ::-1]

    vptree_distances2 = np.sort(vptree_distances, axis=-1)
    assert np.array_equal(vptree_distances, vptree_distances2)  # distances are sorted

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_large_dataset_highdim(vptree_cls, exaustive_metric):
    dimension = 16
    num_points = 401001
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    num_queries = 8
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    k = 3

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float32)[:, ::-1]

    vptree_distances2 = np.sort(vptree_distances, axis=-1)
    assert np.array_equal(vptree_distances, vptree_distances2)  # distances are sorted

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_dataset_split_less_than_k(vptree_cls, exaustive_metric):
    """doc
    Test the case where on of the splits of the dataset eliminates half of points but contains less than k points
    """

    data = np.array([[-2.5, 0], [-2.58, 0], [0, 0], [2.5, 0], [2.6, 0]])
    queries = np.array([[-2.55, 0]])
    k = 4

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float32)[:, ::-1]

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_query_larger_than_dataset(vptree_cls, exaustive_metric):
    num_points = 5
    dimension = 8
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)
    print(f"data: {data}")

    num_queries = 8
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)
    print(f"queries: {queries}")

    k = 3

    exaustive_indices, exaustive_distances = exaustive_metric(data, np.array([queries[0]]), k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(np.array([queries[0]]), k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float32)[:, ::-1]

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_compare_with_exaustive_knn(vptree_cls, exaustive_metric):
    num_points = 21231
    dimension = 8
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    num_queries = 23
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    k = 3

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, k)

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.searchKNN(queries, k)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)[:, ::-1]
    vptree_distances = np.array(vptree_distances, dtype=np.float32)[:, ::-1]

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)


@pytest.mark.parametrize("vptree_cls, exaustive_metric", CLASSES)
def test_compare_with_exaustive_1nn(vptree_cls, exaustive_metric):
    num_points = 21231
    dimension = 8
    data = np.random.rand(num_points, dimension).astype(dtype=np.float32)

    num_queries = 23
    queries = np.random.rand(num_queries, dimension).astype(dtype=np.float32)

    exaustive_indices, exaustive_distances = exaustive_metric(data, queries, 1)

    exaustive_indices = exaustive_indices.reshape(exaustive_indices.shape[:-1])
    exaustive_distances = exaustive_distances.reshape(exaustive_distances.shape[:-1])

    vptree = vptree_cls()
    vptree.set(data)
    vptree_indices, vptree_distances = vptree.search1NN(queries)

    vptree_indices = np.array(vptree_indices, dtype=np.uint64)
    vptree_distances = np.array(vptree_distances, dtype=np.float32)

    assert np.array_equal(exaustive_indices, vptree_indices)
    np.testing.assert_allclose(exaustive_distances, vptree_distances, rtol=1e-06)
